/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, 1991, 1992, William Cheng.
 * 
 * Permission limited to the use, copy, modify, and distribute this software
 * and its documentation for any purpose is hereby granted by the Author without
 * fee, provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the Author not be used
 * in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The Author makes no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.  All other
 * rights are reserved by the Author.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
 * USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /amnt/kona/tangram/u/william/X11/TGIF2/RCS/choice.c,v 2.56 1992/06/19 18:10:16 william Exp $";
#endif

#include <X11/Xlib.h>
#include "const.h"
#include "types.h"

#include "align.e"
#include "arc.e"
#include "box.e"
#include "color.e"
#include "cursor.e"
#include "font.e"
#include "grid.e"
#include "mainmenu.e"
#include "mark.e"
#include "msg.e"
#include "oval.e"
#include "pattern.e"
#include "poly.e"
#include "polygon.e"
#include "raster.e"
#include "rcbox.e"
#include "select.e"
#include "setup.e"
#include "text.e"

#define MODE_ROW 0
#define EDIT_ROW 0
#define ZOOM_ROW 0
#define RADIUS_ROW 0
#define HORI_ALIGN_ROW 0
#define FONT_ROW 0
#define VSPACE_ROW 0
#define DASH_ROW 0
#define LINE_TYPE_ROW 0
#define FILL_ROW 0
#define COLOR_ROW 0

#define FILE_ROW 1
#define PRINT_ROW 1
#define MOVE_MODE_ROW 1
#define ROTATE_ROW 1
#define VERT_ALIGN_ROW 1
#define TEXT_SIZE_ROW 1
#define JUST_ROW 1
#define LINE_STYLE_ROW 1
#define LINE_WIDTH_ROW 1
#define PEN_ROW 1
#define SPECIAL_ROW 1

#define MODE_COL 0
#define EDIT_COL 1
#define ZOOM_COL 2
#define RADIUS_COL 3
#define HORI_ALIGN_COL 4
#define FONT_COL 5
#define VSPACE_COL 6
#define DASH_COL 7
#define LINE_TYPE_COL 8
#define FILL_COL 9
#define COLOR_COL 10

#define FILE_COL 0
#define PRINT_COL 1
#define MOVE_MODE_COL 2
#define ROTATE_COL 3
#define VERT_ALIGN_COL 4
#define TEXT_SIZE_COL 5
#define JUST_COL 6
#define LINE_STYLE_COL 7
#define LINE_WIDTH_COL 8
#define PEN_COL 9
#define SPECIAL_COL 10

#define CHOICE_SHIFT 4

#define CHOICE_MODE ((MODE_COL<<CHOICE_SHIFT)|MODE_ROW)
#define CHOICE_COLOR ((COLOR_COL<<CHOICE_SHIFT)|COLOR_ROW)
#define CHOICE_FILE ((FILE_COL<<CHOICE_SHIFT)|FILE_ROW)
#define CHOICE_EDIT ((EDIT_COL<<CHOICE_SHIFT)|EDIT_ROW)
#define CHOICE_PRINT ((PRINT_COL<<CHOICE_SHIFT)|PRINT_ROW)
#define CHOICE_SPECIAL ((SPECIAL_COL<<CHOICE_SHIFT)|SPECIAL_ROW)
#define CHOICE_HORI_ALIGN ((HORI_ALIGN_COL<<CHOICE_SHIFT)|HORI_ALIGN_ROW)
#define CHOICE_VERT_ALIGN ((VERT_ALIGN_COL<<CHOICE_SHIFT)|VERT_ALIGN_ROW)
#define CHOICE_ROTATE ((ROTATE_COL<<CHOICE_SHIFT)|ROTATE_ROW)
#define CHOICE_MOVE_MODE ((MOVE_MODE_COL<<CHOICE_SHIFT)|MOVE_MODE_ROW)
#define CHOICE_RADIUS ((RADIUS_COL<<CHOICE_SHIFT)|RADIUS_ROW)
#define CHOICE_ZOOM ((ZOOM_COL<<CHOICE_SHIFT)|ZOOM_ROW)
#define CHOICE_FONT ((FONT_COL<<CHOICE_SHIFT)|FONT_ROW)
#define CHOICE_JUST ((JUST_COL<<CHOICE_SHIFT)|JUST_ROW)
#define CHOICE_TEXT_SIZE ((TEXT_SIZE_COL<<CHOICE_SHIFT)|TEXT_SIZE_ROW)
#define CHOICE_VSPACE ((VSPACE_COL<<CHOICE_SHIFT)|VSPACE_ROW)
#define CHOICE_LINE_WIDTH ((LINE_WIDTH_COL<<CHOICE_SHIFT)|LINE_WIDTH_ROW)
#define CHOICE_LINE_STYLE ((LINE_STYLE_COL<<CHOICE_SHIFT)|LINE_STYLE_ROW)
#define CHOICE_LINE_TYPE ((LINE_TYPE_COL<<CHOICE_SHIFT)|LINE_TYPE_ROW)
#define CHOICE_DASH ((DASH_COL<<CHOICE_SHIFT)|DASH_ROW)
#define CHOICE_FILL ((FILL_COL<<CHOICE_SHIFT)|FILL_ROW)
#define CHOICE_PEN ((PEN_COL<<CHOICE_SHIFT)|PEN_ROW)

int	curChoice = NOTHING;

static GC	choiceGC;
static Pixmap	choiceBackingPixmap;

extern void	ShowWhereToPrint ();

void InitChoice ()
{
   XGCValues	values;

   values.foreground = xorOne;
   values.background = xorZero;
   values.fill_style = FillSolid;
#ifdef XI18N
   choiceGC = XCreateGC (mainDisplay, choiceWindow,
         GCForeground | GCBackground | GCFillStyle, &values);
#else
   values.font = defaultFontPtr->fid;
   choiceGC = XCreateGC (mainDisplay, choiceWindow,
         GCForeground | GCBackground | GCFillStyle | GCFont, &values);
#endif

   choiceBackingPixmap = XCreatePixmap (mainDisplay, choiceWindow,
         choiceWindowW, choiceWindowH, mainDepth);
}

void CleanUpChoices ()
{
   XFreeGC (mainDisplay, choiceGC);
   XFreePixmap (mainDisplay, choiceBackingPixmap);
}

void ShowMode ()
{
   XSetStipple (mainDisplay, rasterGC, choicePixmap[curChoice]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, MODE_COL*choiceImageW,
         MODE_ROW*choiceImageH, choiceImageW, choiceImageH);
}

void ShowColor (PropagateWhereToPrint)
   int	PropagateWhereToPrint;
{
   XGCValues	values;

   if (colorDisplay)
   {
      values.foreground = colorPixels[colorIndex];
      values.function = GXcopy;
      values.fill_style = FillOpaqueStippled;
      values.stipple = patPixmap[1];
      XChangeGC (mainDisplay, patGC,
            GCForeground | GCFunction | GCFillStyle | GCStipple, &values);

      XFillRectangle (mainDisplay, choiceWindow, patGC,
            COLOR_COL*choiceImageW, COLOR_ROW*choiceImageH,
            choiceImageW, choiceImageH);
   }
   if (PropagateWhereToPrint) ShowWhereToPrint ();
}

void ShowHoriAlign ()
{
   XSetStipple (mainDisplay, rasterGC, alignHoriPixmap[horiAlign]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC,
         HORI_ALIGN_COL*choiceImageW, HORI_ALIGN_ROW*choiceImageH,
         choiceImageW, choiceImageH);
}

void ShowVertAlign ()
{
   XSetStipple (mainDisplay, rasterGC, alignVertPixmap[vertAlign]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC,
         VERT_ALIGN_COL*choiceImageW, VERT_ALIGN_ROW*choiceImageH,
         choiceImageW, choiceImageH);
}

void ShowJust ()
{
   XSetStipple (mainDisplay, rasterGC, justPixmap[textJust]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC,
         JUST_COL*choiceImageW, JUST_ROW*choiceImageH,
         choiceImageW, choiceImageH);
}

void ShowCurFont ()
{
   int		x, y, w, h;
   XRectangle	recs[1];
   XGCValues	values;

   recs[0].x = FONT_COL*choiceImageW;
   recs[0].y = FONT_ROW*choiceImageH;
   recs[0].width = choiceImageW;
   recs[0].height = choiceImageH;

#ifdef XI18N
   if (curFont == FONT_MUL)
      w = MyTextWidth (canvasFontPtr, canvasFontSymPtr, HIRAGANA_A_EUC,
			HIRAGANA_A_LEN);
   else
      w = MyTextWidth (canvasFontPtr, canvasFontSymPtr, "W", 1);
#else
   w = XTextWidth (canvasFontPtr, "W", 1);
#endif
   h = canvasFontHeight;

   x = (w > choiceImageW) ?
         FONT_COL*choiceImageW - (w-choiceImageW)/2 :
         FONT_COL*choiceImageW + (choiceImageW-w)/2;
   y = (h > choiceImageH) ?
         FONT_ROW*choiceImageH + (canvasFontAsc-(h-choiceImageH)/2) :
         FONT_ROW*choiceImageH + (canvasFontAsc+(choiceImageH-h)/2);

   XClearArea (mainDisplay, choiceWindow, FONT_COL*choiceImageW,
         FONT_ROW*choiceImageH, choiceImageW, choiceImageH, FALSE);

#ifdef XI18N
   if (curFont == FONT_SYM)
      XSetFont (mainDisplay, choiceGC, canvasFontSymPtr->fid);
#else
   XSetFont (mainDisplay, choiceGC, canvasFontPtr->fid);
#endif
   XSetClipRectangles (mainDisplay, choiceGC, 0, 0, recs, 1, YXBanded);

   values.foreground = myFgPixel;
   values.background = myBgPixel;
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground, &values);

#ifdef XI18N
   if (curFont == FONT_MUL)
       XmbDrawString (mainDisplay, choiceWindow, canvasFontPtr, choiceGC,
		x, y, HIRAGANA_A_EUC, HIRAGANA_A_LEN);
   else if (curFont != FONT_SYM)
       XmbDrawString (mainDisplay, choiceWindow, canvasFontPtr, choiceGC,
		x, y, "W", 1);
   else 
#endif
   XDrawString (mainDisplay, choiceWindow, choiceGC, x, y, "W", 1);

   values.foreground = xorOne;
   values.background = xorZero;
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground, &values);

   recs[0].x = 0;
   recs[0].y = 0;
   recs[0].width = choiceWindowW;
   recs[0].height = choiceWindowH;
   XSetClipRectangles (mainDisplay, choiceGC, 0, 0, recs, 1, YXBanded);
}

void ShowTextVSpace ()
{
   int		len, x, y, w;
   char		s[80];
   XGCValues	values;

   XSetStipple (mainDisplay, rasterGC, vspacePixmap);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, VSPACE_COL*choiceImageW,
         VSPACE_ROW*choiceImageH, choiceImageW, choiceImageH);
   sprintf (s, "%1d", textVSpace);
   len = strlen (s);
#ifdef XI18N
   w = XmbTextEscapement (rulerFontPtr, s, len);
#else
   w = XTextWidth (rulerFontPtr, s, len);
#endif
   x = (int)((VSPACE_COL+0.5)*choiceImageW-2);
   y = VSPACE_ROW*choiceImageH+((choiceImageH-rulerFontAsc)>>1)+rulerFontAsc;

   values.foreground = myFgPixel;
   values.background = myBgPixel;
#ifdef XI18N
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground,
         &values);

   XmbDrawString (mainDisplay, choiceWindow, rulerFontPtr, choiceGC, x, y, s, len);
#else
   values.font = rulerFontPtr->fid;
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground | GCFont,
         &values);

   XDrawString (mainDisplay, choiceWindow, choiceGC, x, y, s, len);
#endif

   values.foreground = xorOne;
   values.background = xorZero;
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground, &values);
}

void ShowZoom ()
{
   int		len, x, y, w, x_w, one_w;
   char		s[80];
   XGCValues	values;

   XClearArea (mainDisplay, choiceWindow, ZOOM_COL*choiceImageW,
      ZOOM_ROW*choiceImageH, choiceImageW, choiceImageH, FALSE);

   sprintf (s, "x%1d", 1<<zoomScale);
   len = strlen (s);

#ifdef XI18N
   w = XmbTextEscapement (rulerFontPtr, s, len)+1;
   x_w = XmbTextEscapement (rulerFontPtr, "x", 1)+1;
   one_w = XmbTextEscapement (rulerFontPtr, "1", 1);
#else
   w = XTextWidth (rulerFontPtr, s, len)+1;
   x_w = XTextWidth (rulerFontPtr, "x", 1)+1;
   one_w = XTextWidth (rulerFontPtr, "1", 1);
#endif
   x = ZOOM_COL*choiceImageW+((choiceImageW-w)>>1);
   y = ZOOM_ROW*choiceImageH+((choiceImageH-rulerFontAsc)>>1)+rulerFontAsc;

   values.foreground = myFgPixel;
   values.background = myBgPixel;
#ifdef XI18N
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground,
         &values);

   XmbDrawString (mainDisplay, choiceWindow, rulerFontPtr, choiceGC, x, y, "x", 1);
#else
   values.font = rulerFontPtr->fid;
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground | GCFont,
         &values);

   XDrawString (mainDisplay, choiceWindow, choiceGC, x, y, "x", 1);
#endif

   x += x_w;
   sprintf (s, "%1d", 1<<zoomScale);
   len--;
   if (zoomedIn || zoomScale==0)
      XDrawString (mainDisplay, choiceWindow, choiceGC, x, y, s, len);
   else
   {
      y = ZOOM_ROW*choiceImageH + (choiceImageH>>1);
      XDrawString (mainDisplay, choiceWindow, choiceGC, x+((w-x_w-one_w)>>1),
            y-2, "1", 1);
      XDrawLine (mainDisplay, choiceWindow, choiceGC, x, y, x+(w-x_w-1), y);
      XDrawString (mainDisplay, choiceWindow, choiceGC, x, y+rulerFontAsc,
            s, len);
   }

   values.foreground = xorOne;
   values.background = xorZero;
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground, &values);
}

void ShowTextSize ()
{
   int		len, x, y, w;
   char		s[80];
   XGCValues	values;

   XClearArea (mainDisplay, choiceWindow, TEXT_SIZE_COL*choiceImageW,
         TEXT_SIZE_ROW*choiceImageH, choiceImageW, choiceImageH, FALSE);
   switch (curFontDPI)
   {
      case FONT_DPI_75: sprintf (s, "%1d", pointSize75[curSize]); break;
      case FONT_DPI_100: sprintf (s, "%1d", pointSize100[curSize]); break;
   }
   len = strlen (s);
#ifdef XI18N
   w = XmbTextEscapement (defaultFontPtr, s, len);
#else
   w = XTextWidth (defaultFontPtr, s, len);
#endif
   x = TEXT_SIZE_COL*choiceImageW + ((choiceImageW-w)>>1);
   y = TEXT_SIZE_ROW*choiceImageH + ((choiceImageH-defaultFontAsc)>>1) + 
         defaultFontAsc;

   values.foreground = myFgPixel;
   values.background = myBgPixel;
#ifdef XI18N
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground,
         &values);

   XmbDrawString (mainDisplay, choiceWindow, defaultFontPtr, choiceGC, x, y, s, len);
#else
   values.font = defaultFontPtr->fid;
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground | GCFont,
         &values);

   XDrawString (mainDisplay, choiceWindow, choiceGC, x, y, s, len);
#endif

   values.foreground = xorOne;
   values.background = xorZero;
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground, &values);
}

void ShowRotate ()
{
   XSetStipple (mainDisplay, rasterGC, rotatePixmap[curRotate]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, ROTATE_COL*choiceImageW,
         ROTATE_ROW*choiceImageH, choiceImageW, choiceImageH);
}

void ShowSpecial ()
{
   XSetStipple (mainDisplay, rasterGC, specialPixmap);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC,
         SPECIAL_COL*choiceImageW, SPECIAL_ROW*choiceImageH,
         choiceImageW, choiceImageH);
}

void ShowLineWidth ()
{
   XSetStipple (mainDisplay, rasterGC, shortLineWidthPixmap[lineWidth]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC,
         LINE_WIDTH_COL*choiceImageW, LINE_WIDTH_ROW*choiceImageH,
         choiceImageW, choiceImageH);
}

void ShowLineStyle ()
{
   XSetStipple (mainDisplay, rasterGC, shortLineStylePixmap[lineStyle]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC,
         LINE_STYLE_COL*choiceImageW, LINE_STYLE_ROW*choiceImageH,
         choiceImageW, choiceImageH);
}

void ShowLineType ()
{
   XSetStipple (mainDisplay, rasterGC, shortLineTypePixmap[curSpline]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC,
         LINE_TYPE_COL*choiceImageW, LINE_TYPE_ROW*choiceImageH,
         choiceImageW, choiceImageH);
}

void ShowDash ()
{
   XSetStipple (mainDisplay, rasterGC, shortDashPixmap[curDash]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, DASH_COL*choiceImageW,
         DASH_ROW*choiceImageH, choiceImageW, choiceImageH);
}

void ShowWhereToPrint ()
{
   if (colorDump)
      XSetForeground (mainDisplay, rasterGC, colorPixels[colorIndex]);
   XSetStipple (mainDisplay, rasterGC, whereToPrintPixmap[whereToPrint]);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, PRINT_COL*choiceImageW,
         PRINT_ROW*choiceImageH, choiceImageW, choiceImageH);
   if (colorDump) XSetForeground (mainDisplay, rasterGC, myFgPixel);
}

void ShowFile ()
{
   switch (pageStyle)
   {
      case PORTRAIT:
         XSetStipple (mainDisplay, rasterGC, filePixmap);
         XFillRectangle (mainDisplay, choiceWindow, rasterGC,
               FILE_COL*choiceImageW, FILE_ROW*choiceImageH,
               choiceImageW, choiceImageH);
         break;
      case LANDSCAPE:
         XSetStipple (mainDisplay, rasterGC, landscapePixmap);
         XFillRectangle (mainDisplay, choiceWindow, rasterGC,
               FILE_COL*choiceImageW, FILE_ROW*choiceImageH,
               choiceImageW, choiceImageH);
         break;
   }
}

void ShowEdit ()
{
   XSetStipple (mainDisplay, rasterGC, editPixmap);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, EDIT_COL*choiceImageW,
         EDIT_ROW*choiceImageH, choiceImageW, choiceImageH);
}

void ShowRCBRadius ()
{
   int		len, x, y, w;
   char		s[80];
   XGCValues	values;

   XSetStipple (mainDisplay, rasterGC, rcbRadiusPixmap);
   XFillRectangle (mainDisplay, choiceWindow, rasterGC, RADIUS_COL*choiceImageW,
         RADIUS_ROW*choiceImageH, choiceImageW, choiceImageH);
   sprintf (s, "%1d", rcbRadius);
   len = strlen (s);
#ifdef XI18N
   w = XmbTextEscapement (rulerFontPtr, s, len);
#else
   w = XTextWidth (rulerFontPtr, s, len);
#endif
   x = (int)((RADIUS_COL+0.5)*choiceImageW-2);
   y = (((RADIUS_ROW+1)*choiceImageH-rulerFontAsc)>>1)+rulerFontAsc;

   values.foreground = myFgPixel;
   values.background = myBgPixel;
#ifdef XI18N
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground,
         &values);

   XmbDrawString (mainDisplay, choiceWindow, rulerFontPtr, choiceGC, x, y, s, len);
#else
   values.font = rulerFontPtr->fid;
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground | GCFont,
         &values);

   XDrawString (mainDisplay, choiceWindow, choiceGC, x, y, s, len);
#endif

   values.foreground = xorOne;
   values.background = xorZero;
   XChangeGC (mainDisplay, choiceGC, GCForeground | GCBackground, &values);
}

void ShowMoveMode ()
{
   switch (moveMode)
   {
      case CONST_MOVE:
         XSetStipple (mainDisplay, rasterGC, moveModePixmap[CONST_MOVE]);
         XFillRectangle (mainDisplay, choiceWindow, rasterGC,
               MOVE_MODE_COL*choiceImageW, MOVE_MODE_ROW*choiceImageH,
               choiceImageW, choiceImageH);
         break;
      case UNCONST_MOVE:
         XSetStipple (mainDisplay, rasterGC, moveModePixmap[UNCONST_MOVE]);
         XFillRectangle (mainDisplay, choiceWindow, rasterGC,
               MOVE_MODE_COL*choiceImageW, MOVE_MODE_ROW*choiceImageH,
               choiceImageW, choiceImageH);
         break;
   }
}

void ShowFill ()
{
   XGCValues	values;

   values.foreground = myFgPixel;
   values.background = myBgPixel;
   values.function = GXcopy;
   values.fill_style = FillOpaqueStippled;
   values.stipple = patPixmap[objFill];
   XChangeGC (mainDisplay, patGC,
         GCForeground | GCBackground | GCFunction | GCFillStyle | GCStipple,
         &values);

   XFillRectangle (mainDisplay, choiceWindow, patGC, FILL_COL*choiceImageW,
         FILL_ROW*choiceImageH, choiceImageW, choiceImageH);
}

void ShowPen ()
{
   XGCValues	values;

   values.foreground = myFgPixel;
   values.background = myBgPixel;
   values.function = GXcopy;
   values.fill_style = FillOpaqueStippled;
   values.stipple = patPixmap[penPat];
   XChangeGC (mainDisplay, patGC,
         GCForeground | GCBackground | GCFunction | GCFillStyle | GCStipple,
         &values);

   XFillRectangle (mainDisplay, choiceWindow, patGC, PEN_COL*choiceImageW,
         PEN_ROW*choiceImageH, choiceImageW, choiceImageH);
   if (penPat != NONEPAT)
      XClearArea (mainDisplay, choiceWindow,
            PEN_COL*choiceImageW+(choiceImageW>>2),
            PEN_ROW*choiceImageH+(choiceImageH>>2),
            (choiceImageW>>1), (choiceImageH>>1), FALSE);
}

void SetCurChoice (NewChoice)
   int	NewChoice;
{
   if (curChoice == NewChoice) return;

   switch (curChoice)
   {
      case NOTHING:
         if (topSel != NULL)
         {
            if (NewChoice == VERTEXMODE)
            {
               UnSelNonVertexObjs (TRUE); /* with highlight */
               UpdSelBBox ();
            }
            else
            {
               HighLightReverse ();
               RemoveAllSel ();
            }
         }
         break;
      case DRAWTEXT:
         CreateTextObj ();
         if (NewChoice == NOTHING && textDrawn) HighLightJustDrawnText ();
         textDrawn = FALSE;
         textCursorShown = FALSE;
         break;
      case DRAWBOX:
         if (NewChoice == NOTHING && boxDrawn) SelectTopObj ();
         boxDrawn = FALSE;
         break;
      case DRAWCIRCLE:
         if (NewChoice == NOTHING && ovalDrawn) SelectTopObj ();
         ovalDrawn = FALSE;
         break;
      case DRAWPOLY:
         if ((NewChoice==NOTHING || NewChoice==VERTEXMODE) && polyDrawn)
            SelectTopObj ();
         polyDrawn = FALSE;
         break;
      case DRAWPOLYGON:
         if ((NewChoice == NOTHING || NewChoice==VERTEXMODE) && polygonDrawn)
            SelectTopObj ();
         polygonDrawn = FALSE;
         break;
      case DRAWARC:
         if (NewChoice == NOTHING && arcDrawn) SelectTopObj ();
         arcDrawn = FALSE;
         break;
      case DRAWRCBOX:
         if (NewChoice == NOTHING && rcBoxDrawn) SelectTopObj ();
         rcBoxDrawn = FALSE;
         break;
      case FREEHAND:
         if ((NewChoice==NOTHING || NewChoice==VERTEXMODE) && polyDrawn)
            SelectTopObj ();
         polyDrawn = FALSE;
         break;
      case VERTEXMODE:
         if (NewChoice == NOTHING)
         {
            HighLightReverse ();
            JustRemoveAllVSel ();
            HighLightForward ();
         }
         else
         {
            HighLightReverse ();
            RemoveAllSel ();
         }
         break;
   }

   curChoice = NewChoice;

   ShowCursor ();
   textCursorShown = FALSE;

   ShowMode ();
   UpdateSubMenu (MENU_MODE);
}

static int	prevChoice = NOTHING;

void PushCurChoice ()
   /* Kouichi Matsuda's modification */
{
   if (curChoice == NOTHING)
      SetCurChoice (prevChoice);
   else
   {
      prevChoice = curChoice;
      SetCurChoice (NOTHING);
   }
}

static
void TextVSpaceLoop (button_ev)
   XButtonEvent	* button_ev;
{
   int		saved_text_vspace = textVSpace, res=8, done=FALSE;
   int		orig_x=button_ev->x, orig_y=button_ev->y;
   int		change=0, saved_change=0;
   XEvent	ev;

   XGrabPointer (mainDisplay, choiceWindow, False,
         PointerMotionMask | ButtonReleaseMask, GrabModeAsync,
         GrabModeAsync, None, handCursor, CurrentTime);

   while (!done)
   {
      XNextEvent (mainDisplay, &ev);

      if (ev.type == Expose || ev.type == VisibilityNotify)
         ExposeEventHandler (&ev, TRUE);
      else if (ev.type == ButtonRelease)
      {
         XUngrabPointer (mainDisplay, CurrentTime);
         done = TRUE;
      }
      else if (ev.type == MotionNotify)
      {
         int	dx = ev.xmotion.x - orig_x;
         int	dy = ev.xmotion.y - orig_y;

         change = (abs(dx)>abs(dy)) ? (int)(dx/res) : (int)(dy/res);

         if (change != saved_change)
         {
            textVSpace = saved_text_vspace+change;
            ShowTextVSpace ();
            saved_change = change;
         }
         while (XCheckMaskEvent (mainDisplay, PointerMotionMask, &ev)) ;
      }
   }
   if (topSel == NULL)
   {
      if (textCursorH+textVSpace <= 0)
      {
         Msg ("Text vertical spacing too small.  No change.");
         textVSpace = saved_text_vspace;
         ShowTextVSpace ();
      }
      else
      {
         int	text_cursor_shown = textCursorShown;
         int	text_obj_created, new_text_vspace = textVSpace;

         textVSpace = saved_text_vspace;
         text_obj_created = TieLooseEnds ();
         if (!text_obj_created && curChoice == DRAWTEXT && text_cursor_shown)
         {
            NewCurText ();
            RedrawCurText ();
         }
         else
            textCursorShown = FALSE;
         textVSpace = new_text_vspace;
         ShowTextVSpace ();
      }
   }
   else
   {
      ShowTextVSpace ();
      ChangeVSpace (textVSpace);
   }
}

static
void RCBRadiusLoop (button_ev)
   XButtonEvent	* button_ev;
{
   int		saved_rcb_radius = rcbRadius, res=8, done=FALSE;
   int		orig_x=button_ev->x, orig_y=button_ev->y;
   int		change=0, saved_change=0;
   XEvent	ev;

   XGrabPointer (mainDisplay, choiceWindow, False,
         PointerMotionMask | ButtonReleaseMask, GrabModeAsync,
         GrabModeAsync, None, handCursor, CurrentTime);

   while (!done)
   {
      XNextEvent (mainDisplay, &ev);

      if (ev.type == Expose || ev.type == VisibilityNotify)
         ExposeEventHandler (&ev, TRUE);
      else if (ev.type == ButtonRelease)
      {
         XUngrabPointer (mainDisplay, CurrentTime);
         done = TRUE;
      }
      else if (ev.type == MotionNotify)
      {
         int	dx = ev.xmotion.x - orig_x;
         int	dy = ev.xmotion.y - orig_y;

         change = (abs(dx)>abs(dy)) ? (int)(dx/res) : (int)(dy/res);

         if (change != saved_change)
         {
            rcbRadius = max(MIN_RCB_RADIUS, saved_rcb_radius+change);
            ShowRCBRadius ();
            saved_change = change;
         }
         while (XCheckMaskEvent (mainDisplay, PointerMotionMask, &ev)) ;
      }
   }
   if (topSel != NULL) ChangeAllSelRCBRadius (rcbRadius);
}

static
void ZoomLoop (button_ev)
   XButtonEvent	* button_ev;
{
   int		saved_zoomed_in=zoomedIn, saved_zoom_scale=zoomScale, res=8;
   int		new_zoomed_in, new_zoom_scale;
   int		initial_zoom, zoom, max_zoom, win_w, win_h;
   int		orig_x=button_ev->x, orig_y=button_ev->y;
   int		done=FALSE, change=0, saved_change=0;
   XEvent	ev;

   initial_zoom = (zoomedIn) ? (MAX_ZOOMED_IN-zoomScale) :
         (MAX_ZOOMED_IN+zoomScale);

   win_w = drawWinW;
   win_h = drawWinH;
   new_zoomed_in = zoomedIn;
   new_zoom_scale = zoomScale;
   while ((win_w>>1) >= paperWidth && (win_h>>1) >= paperHeight)
   {
      if (new_zoomed_in)
         new_zoom_scale++;
      else if (new_zoom_scale == 0)
      {
         new_zoomed_in = TRUE;
         new_zoom_scale++;
      }
      else
         new_zoom_scale--;
      win_w >>= 1;
      win_h >>= 1;
   }
   while (win_w < paperWidth || win_h < paperHeight)
   {
      if (!new_zoomed_in)
         new_zoom_scale++;
      else if (new_zoom_scale == 1)
      {
         new_zoomed_in = FALSE;
         new_zoom_scale--;
      }
      else
         new_zoom_scale--;
      win_w <<= 1;
      win_h <<= 1;
   }
   max_zoom = (new_zoomed_in) ? (MAX_ZOOMED_IN-new_zoom_scale) :
         (MAX_ZOOMED_IN+new_zoom_scale);

   XGrabPointer (mainDisplay, choiceWindow, False,
         PointerMotionMask | ButtonReleaseMask, GrabModeAsync,
         GrabModeAsync, None, handCursor, CurrentTime);

   while (!done)
   {
      XNextEvent (mainDisplay, &ev);

      if (ev.type == Expose || ev.type == VisibilityNotify)
         ExposeEventHandler (&ev, TRUE);
      else if (ev.type == ButtonRelease)
      {
         XUngrabPointer (mainDisplay, CurrentTime);
         done = TRUE;
      }
      else if (ev.type == MotionNotify)
      {
         int	dx = ev.xmotion.x - orig_x;
         int	dy = ev.xmotion.y - orig_y;

         change = (abs(dx)>abs(dy)) ? (int)(dx/res) : (int)(dy/res);

         if (change != saved_change)
         {
            zoom = initial_zoom+change;
            if (zoom < 0) zoom = 0;
            if (zoom > max_zoom) zoom = max_zoom;
            zoomedIn = (zoom < MAX_ZOOMED_IN);
            zoomScale = (zoomedIn) ? (MAX_ZOOMED_IN-zoom) :
                  (zoom-MAX_ZOOMED_IN);
            ShowZoom ();
            saved_change = change;
         }
         while (XCheckMaskEvent (mainDisplay, PointerMotionMask, &ev)) ;
      }
   }
   if (saved_zoomed_in==zoomedIn && saved_zoom_scale==zoomScale) return;
   new_zoomed_in = zoomedIn;
   new_zoom_scale = zoomScale;
   zoomedIn = saved_zoomed_in;
   zoomScale = saved_zoom_scale;
   PreciseZoom (new_zoomed_in, new_zoom_scale);
}

int ChoiceEventHandler (input)
   XEvent	* input;
{
   XEvent	ev;
   XButtonEvent	* button_ev;
   int		xindex, yindex, delta=0, root_x, root_y, x=0, y=0;
   int		choice=INVALID;
   Window	root_win, child_win;
   unsigned int	status;

   if (input->type == Expose)
   {
      XSync (mainDisplay, False);
      while (XCheckWindowEvent (mainDisplay, choiceWindow, ExposureMask, &ev)) ;
      RedrawChoiceWindow ();
   }
   else if (input->type == ButtonPress)
   {
      Msg ("");
      button_ev = &(input->xbutton);
      switch (button_ev->button)
      {
         case Button1: delta = 1; break;
         case Button2:
            XQueryPointer (mainDisplay, rootWindow, &root_win, &child_win,
                  &root_x, &root_y, &x, &y, &status);
            break;
         case Button3: delta = -1; break;
         default: return (INVALID);
      }
      xindex = (int)(button_ev->x / choiceImageW);
      yindex = (int)(button_ev->y / choiceImageH);
      choice = (xindex<<CHOICE_SHIFT)|yindex;
      switch (choice)
      {
         case CHOICE_MODE:
            if (delta == 0)
               ModeMenu (x, y);
            else
               SetCurChoice ((curChoice+MAXCHOICES+delta) % MAXCHOICES);
            break;
         case CHOICE_COLOR:
            if (colorDisplay)
            {
               if (delta == 0)
                  ColorMenu (x, y);
               else
               {
                  if (curChoice != NOTHING) SetCurChoice (NOTHING);
                  colorIndex = (colorIndex+maxColors+delta) % maxColors;
                  ShowColor (TRUE);
                  if (topSel != NULL) ChangeAllSelColor (colorIndex);
               }
            }
            break;
         case CHOICE_HORI_ALIGN:
            if (delta == 0)
               HoriAlignMenu (x, y);
            else
               HoriAlignSubMenu ((horiAlign+MAXALIGNS+delta) % MAXALIGNS);
            break;
         case CHOICE_VERT_ALIGN:
            if (delta == 0)
               VertAlignMenu (x, y);
            else
               VertAlignSubMenu ((vertAlign+MAXALIGNS+delta) % MAXALIGNS);
            break;
         case CHOICE_JUST:
            if (delta == 0)
               StyleMenu (x, y);
            else
            {
               if (curChoice == DRAWTEXT)
               {
                  if (TieLooseEnds ())
                  {
                     SetCurChoice (NOTHING);
                     HighLightReverse ();
                  }
                  else
                     SetCurChoice (NOTHING);
                  RemoveAllSel ();
               }
               textJust = (textJust+MAXJUSTS+delta) % MAXJUSTS;
               ShowJust ();
               UpdateSubMenu (MENU_STYLE);
               if (topSel != NULL) ChangeFontJust (textJust);
            }
            break;
         case CHOICE_FONT:
            if (delta == 0)
               FontMenu (x, y);
            else
            {
#ifdef XI18N
               /*
                * can not exchange Multibyte font set with one font
                * (English or Symbol).
                */
               if (ConflictTextFont((curFont+MAXFONTS+delta) % MAXFONTS)) {
                  Msg("Conflict with multibyte font.");
                  break;
               }
#endif
               if (curChoice == DRAWTEXT)
               {
                  if (TieLooseEnds ())
                  {
                     SetCurChoice (NOTHING);
                     HighLightReverse ();
                  }
                  else
                     SetCurChoice (NOTHING);
                  RemoveAllSel ();
               }
               curFont = (curFont+MAXFONTS+delta) % MAXFONTS;
#ifdef XI18N
               if (curFont == FONT_SYM)
                  SetCanvasFontSym ();
               else
#endif
               SetCanvasFont ();
               ShowCurFont ();
               ShowTextSize ();
               UpdateSubMenu (MENU_FONT);
               if (topSel != NULL) ChangeFont (curFont);
            }
            break;
         case CHOICE_VSPACE:
            if (delta == 0)
               TextVSpaceLoop (button_ev);
            else
            {
               if (curChoice == DRAWTEXT)
               {
                  if (TieLooseEnds ())
                  {
                     SetCurChoice (NOTHING);
                     HighLightReverse ();
                  }
                  else
                     SetCurChoice (NOTHING);
                  RemoveAllSel ();
               }
               if (topSel == NULL && delta < 0 &&
                     textCursorH+textVSpace+delta <= 0)
                  Msg ("Text vertical spacing too small.  No change.");
               else
               {
                  textVSpace += delta;
                  ShowTextVSpace ();
                  if (topSel != NULL) ChangeVSpace (textVSpace);
               }
            }
            break;
         case CHOICE_TEXT_SIZE:
            if (delta == 0)
               SizeMenu (x, y);
            else
            {
               int	cur_index;

               if (curChoice == DRAWTEXT)
               {
                  if (TieLooseEnds ())
                  {
                     SetCurChoice (NOTHING);
                     HighLightReverse ();
                  }
                  else
                     SetCurChoice (NOTHING);
                  RemoveAllSel ();
               }
               cur_index = GetSizeMenuIndex ();
               cur_index = (cur_index+2*MAXFONTSIZES-1+delta) %
                     (2*MAXFONTSIZES-1);
               curSize = ParseSizeMenuIndex (cur_index, &curFontDPI);
#ifdef XI18N
               if (curFont == FONT_SYM)
                  SetCanvasFontSym ();
#endif
               SetCanvasFont ();
               ShowCurFont ();
               ShowTextSize ();
               UpdateSubMenu (MENU_SIZE);
               if (topSel != NULL) ChangeFontSize (curSize);
            }
            break;
         case CHOICE_ROTATE:
            if (delta == 0)
               ArrangeMenu (x, y);
            else
            {
               if (curChoice == DRAWTEXT)
               {
                  if (TieLooseEnds ())
                  {
                     SetCurChoice (NOTHING);
                     HighLightReverse ();
                  }
                  else
                     SetCurChoice (NOTHING);
                  RemoveAllSel ();
               }
               curRotate = (curRotate+4+delta) % 4;
               ShowRotate ();
               if (topSel != NULL)
               {
                  if (delta == 1)
                     RotateClockWise ();
                  else
                     RotateCounter ();
               }
            }
            break;
         case CHOICE_EDIT: if (delta == 0) EditMenu (x, y); break;
         case CHOICE_SPECIAL: if (delta == 0) SpecialMenu (x, y); break;
         case CHOICE_LINE_WIDTH:
            if (delta == 0)
               LineWidthMenu (x, y);
            else
            {
               lineWidth = (lineWidth+maxLineWidths+delta) % maxLineWidths;
               ShowLineWidth ();
               UpdateSubMenu (MENU_LINEWIDTH);
               if (topSel != NULL) ChangeAllSelLineWidth (lineWidth);
            }
            break;
         case CHOICE_LINE_STYLE:
            if (delta == 0)
               LineStyleMenu (x, y);
            else
            {
               lineStyle = (lineStyle+MAXLINESTYLES+delta) % MAXLINESTYLES;
               ShowLineStyle ();
               UpdateSubMenu (MENU_LINESTYLE);
               if (topSel != NULL) ChangeAllSelLineStyle (lineStyle);
            }
            break;
         case CHOICE_LINE_TYPE:
            if (delta == 0)
               LineTypeMenu (x, y);
            else
            {
               curSpline = (curSpline+MAXLINETYPES+delta) % MAXLINETYPES;
               ShowLineType ();
               UpdateSubMenu (MENU_LINETYPE);
               if (topSel != NULL) ChangeAllSelLineType (curSpline);
            }
            break;
         case CHOICE_DASH:
            if (delta == 0)
               LineDashMenu (x, y);
            else
            {
               curDash = (curDash+MAXDASHES+delta) % MAXDASHES;
               ShowDash ();
               UpdateSubMenu (MENU_LINEDASH);
               if (topSel != NULL) ChangeAllSelDashes (curDash);
            }
            break;
         case CHOICE_RADIUS:
            if (delta == 0)
               RCBRadiusLoop (button_ev);
            else
            {
               if (rcbRadius+delta >= MIN_RCB_RADIUS)
               {
                  rcbRadius += delta;
                  ShowRCBRadius ();
                  if (topSel != NULL) ChangeAllSelRCBRadius (rcbRadius);
               }
            }
            break;
         case CHOICE_ZOOM:
            switch (delta)
            {
               case -1: ZoomIn (); break;
               case 0: ZoomLoop (button_ev); break;
               case 1: ZoomOut (); break;
            }
            break;
         case CHOICE_MOVE_MODE:
            if (delta == 0)
               MoveModeMenu (x, y);
            else
               ToggleMoveMode ();
            break;
         case CHOICE_PRINT:
            if (delta == 0)
               LayoutMenu (x, y);
            else
            {
               whereToPrint = (whereToPrint+MAXWHERETOPRINT+delta) %
                     MAXWHERETOPRINT-1;
               ToggleWhereToPrint ();
            }
            break;
         case CHOICE_FILE:
            if (delta == 0)
               return (FileMenu (x, y));
            else
            {
               switch (pageStyle)
               {
                  case PORTRAIT:
                     ChangePageStyle (LANDSCAPE, "LandScape"); break;
                  case LANDSCAPE:
                     ChangePageStyle (PORTRAIT, "Portrait"); break;
               }
            }
            break;
         case CHOICE_FILL:
            if (delta == 0)
               FillMenu (x, y);
            else
            {
               if (curChoice == DRAWTEXT)
               {
                  if (TieLooseEnds ())
                  {
                     SetCurChoice (NOTHING);
                     HighLightReverse ();
                  }
                  else
                     SetCurChoice (NOTHING);
                  RemoveAllSel ();
               }
               objFill = (objFill+MAXPATTERNS+delta) % MAXPATTERNS;
               ShowFill ();
               if (topSel != NULL) ChangeAllSelFill (objFill);
            }
            break;
         case CHOICE_PEN:
            if (delta == 0)
               PenMenu (x, y);
            else
            {
               if (curChoice == DRAWTEXT)
               {
                  if (TieLooseEnds ())
                  {
                     SetCurChoice (NOTHING);
                     HighLightReverse ();
                  }
                  else
                     SetCurChoice (NOTHING);
                  RemoveAllSel ();
               }
               penPat = (penPat+MAXPATTERNS+delta) % MAXPATTERNS;
               ShowPen ();
               if (topSel != NULL) ChangeAllSelPen (penPat);
            }
            break;
      }
   }
   return (INVALID);
}
